﻿/**
 * Relative Accuracy Measures for Stroke Gestures (.NET Framework 4.5 C# version)
 *
 * 	    Radu-Daniel Vatavu, Ph.D.
 *	    University Stefan cel Mare of Suceava
 *	    Suceava 720229, Romania
 *	    vatavu@eed.usv.ro
 *
 * 	    Lisa Anthony, Ph.D.
 *      UMBC
 *      Information Systems Department
 *      1000 Hilltop Circle
 *      Baltimore, MD 21250
 *      lanthony@umbc.edu
 *	
 *	    Jacob O. Wobbrock, Ph.D.
 * 	    The Information School
 *	    University of Washington
 *	    Seattle, WA 98195-2840
 *	    wobbrock@uw.edu
 *
 * The academic publication for the relative accuracy measures, and what should be 
 * used to cite it, is:
 *
 *	Vatavu, R.-D., Anthony, L. and Wobbrock, J.O. (2013).  
 *	  Relative accuracy measures for stroke gestures. 
 *	  In Proceedings of the 15th ACM on International Conference 
 *	  on Multimodal Interaction (ICMI '13). ACM, New York, NY, USA, 279-286
 *	  DOI=10.1145/2522848.2522875 http://doi.acm.org/10.1145/2522848.2522875
 *
 * This software is distributed under the "New BSD License" agreement:
 *
 * Copyright (c) 2013, Radu-Daniel Vatavu, Lisa Anthony, and 
 * Jacob O. Wobbrock. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the names of the University Stefan cel Mare of Suceava, 
 *	    University of Washington, nor UMBC, nor the names of its contributors 
 *	    may be used to endorse or promote products derived from this software 
 *	    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Radu-Daniel Vatavu OR Lisa Anthony
 * OR Jacob O. Wobbrock BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
**/
namespace RelativeAccuracyMeasures
{
    /// <summary>
    /// 
    /// Implements the centroid gesture shape of a set of gestures.
    /// 
    /// </summary>
    public class CentroidGesture : Gesture
    {
        public PointAlignmentType PointAlignmentType;   // specifies how points are aligned during the averaging process

        /// <summary>
        /// Constructs the centroid gesture from a set of samples
        /// by aligning points and averaging their X and Y coordinates.
        /// </summary>
        public CentroidGesture(Gesture[] gestures, PointAlignmentType alignmentType = PointAlignmentType.Chronological, bool useNearestNeighbor = true)
            : base(gestures[0].Points, gestures[0].Name, gestures[0].SamplingRate) // initialize the centroid with the first gesture in the set 
                                                                                   // (i.e., the reference to which all the other gesture samples will be aligned)
        {
            this.PointAlignmentType = alignmentType;
            
            // compute the centroid shape of the gesture set
            Point[] centroid = new Point[this.SamplingRate];
            for (int i = 0; i < centroid.Length; i++)
                centroid[i] = new Point();
            foreach (Gesture gesture in gestures)
            {
                Point[] points = AlignGesture(gesture);
                for (int i = 0; i < points.Length; i++)
                    centroid[i] += points[i];
            }
            for (int i = 0; i < centroid.Length; i++)
                centroid[i] = centroid[i] / gestures.Length;
            
            // create the average gesture
            this.OriginalPoints = centroid;
            this.PreprocessGesture(this.SamplingRate);
            
            // if in Nearest-Neighbor mode, the gesture sample that is closest to the centroid becomes the average gesture
            if (useNearestNeighbor)
            {
                centroid = GetPointsForAlignment(this);
                
                int indexNN = -1;
                double min = double.MaxValue;
                for (int indexGesture = 0; indexGesture < gestures.Length; indexGesture++)
                {
                    Point[] points = AlignGesture(gestures[indexGesture]);
                    double distance = 0;
                    for (int i = 0; i < points.Length; i++)
                        distance += Measurements.SqrEuclideanDistance(centroid[i], points[i]);
                    if (distance < min)
                    {
                        min = distance;
                        indexNN = indexGesture;
                    }
                }
                this.OriginalPoints = gestures[indexNN].OriginalPoints;
                this.PreprocessGesture(gestures[indexNN].SamplingRate);
            }
        }

        /// <summary>
        /// Aligns a gesture to the centroid.
        /// </summary>
        public Point[] AlignGesture(Gesture gesture)
        {
            return AlignGesture(gesture, PointAlignmentType);
        }

        public Point[] AlignGesture(Gesture gesture, PointAlignmentType type)
        {
            Point[] points = GetPointsForAlignment(gesture);
            if (type == PointAlignmentType.Chronological)
                return points;
            else
            {
                int[] alignment = null;
                // Hungarian is employed below, but the $P point cloud distance can also be used
                PDollarAlternatives.Hungarian(GetPointsForAlignment(this), points, ref alignment);
                Point[] newPoints = new Point[alignment.Length];
                for (int i = 0; i < alignment.Length; i++)
                    newPoints[i] = points[alignment[i]];
                return newPoints;
            }
        }

        /// <summary>
        /// Returns the points that will be used for the alignment process.
        /// </summary>
        private Point[] GetPointsForAlignment(Gesture gesture)
        {
            Point[] points = gesture.Points;
            return PointSet.TranslateBy(
                points,
                PointSet.Centroid(points)
            );
        }

        /// <summary>
        /// Returns the points of the average gesture
        /// </summary>
        /// <returns></returns>
        public Point[] GetPoints()
        {
            return GetPointsForAlignment(this);
        }
    }
}
