﻿/**
 * Relative Accuracy Measures for Stroke Gestures (.NET Framework 4.5 C# version)
 *
 * 	    Radu-Daniel Vatavu, Ph.D.
 *	    University Stefan cel Mare of Suceava
 *	    Suceava 720229, Romania
 *	    vatavu@eed.usv.ro
 *
 * 	    Lisa Anthony, Ph.D.
 *      UMBC
 *      Information Systems Department
 *      1000 Hilltop Circle
 *      Baltimore, MD 21250
 *      lanthony@umbc.edu
 *	
 *	    Jacob O. Wobbrock, Ph.D.
 * 	    The Information School
 *	    University of Washington
 *	    Seattle, WA 98195-2840
 *	    wobbrock@uw.edu
 *
 * The academic publication for the relative accuracy measures, and what should be 
 * used to cite it, is:
 *
 *	Vatavu, R.-D., Anthony, L. and Wobbrock, J.O. (2013).  
 *	  Relative accuracy measures for stroke gestures. 
 *	  In Proceedings of the 15th ACM on International Conference 
 *	  on Multimodal Interaction (ICMI '13). ACM, New York, NY, USA, 279-286
 *	  DOI=10.1145/2522848.2522875 http://doi.acm.org/10.1145/2522848.2522875
 *
 * This software is distributed under the "New BSD License" agreement:
 *
 * Copyright (c) 2013, Radu-Daniel Vatavu, Lisa Anthony, and 
 * Jacob O. Wobbrock. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the names of the University Stefan cel Mare of Suceava, 
 *	    University of Washington, nor UMBC, nor the names of its contributors 
 *	    may be used to endorse or promote products derived from this software 
 *	    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Radu-Daniel Vatavu OR Lisa Anthony
 * OR Jacob O. Wobbrock BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
**/
using System;

namespace RelativeAccuracyMeasures
{
    /// <summary>
    /// Computes relative accuracy measures for stroke gestures based on the notion of a "gesture task axis".
    /// 
    /// Source: 
    ///         Radu-Daniel Vatavu, Lisa Anthony, and Jacob O. Wobbrock. 2013. 
    ///         Relative accuracy measures for stroke gestures. 
    ///         In Proceedings of the 15th ACM on International conference on multimodal interaction (ICMI '13). ACM, New York, NY, USA, 279-286.
    ///         DOI=10.1145/2522848.2522875 http://doi.acm.org/10.1145/2522848.2522875
    /// 
    /// Inspired by: 
    ///         I. Scott MacKenzie, Tatu Kauppinen, and Miika Silfverberg. 2001. Accuracy measures for evaluating computer pointing devices. 
    ///         In Proceedings of the SIGCHI conference on Human factors in computing systems (CHI '01). ACM, New York, NY, USA, 9-16. 
    ///         http://doi.acm.org/10.1145/365024.365028
    ///         
    /// </summary>
    public class RelativeAccuracy
    {
        #region 6 geometric relative accuracy measures: ShE, ShV, LE, SzE, BE, BV

        /// <summary>
        /// Shape Error (ShE) is defined as the average of the absolute Euclidean distances 
        /// computed between gesture points and task axis points.
        /// </summary>
        public static double ShapeError(Gesture gesture, CentroidGesture taskAxis)
        {
            double[] errors = ComputeLocalShapeErrors(gesture, taskAxis);
            return ComputeAverage(errors);
        }

        /// <summary>
        /// Shape Variability (ShV) is defined as the standard deviation of all point distances to the gesture task axis.
        /// </summary>
        public static double ShapeVariability(Gesture gesture, CentroidGesture taskAxis)
        {
            double[] errors = ComputeLocalShapeErrors(gesture, taskAxis);
            return ComputeStdev(errors);
        }

        private static double[] ComputeLocalShapeErrors(Gesture gesture, CentroidGesture taskAxis)
        {
            // align gesture points with the task axis
            Point[] taskAxisPts = taskAxis.GetPoints();
            Point[] gesturePts = taskAxis.AlignGesture(gesture);

            // compute errors
            double[] errors = new double[gesturePts.Length];
            for (int i = 0; i < gesturePts.Length; i++)
                errors[i] = Measurements.EuclideanDistance(taskAxisPts[i], gesturePts[i]);
            return errors;
        }

        /// <summary>
        /// Length Error (LE) is defined as the absolute difference in path length 
        /// between the gesture and the task axis.
        /// </summary>
        public static double LengthError(Gesture gesture, CentroidGesture taskAxis, bool computeAsPercentage = false)
        {
            double gestureLength = PointSet.PathLength(gesture.Points);
            double taskAxisLength = PointSet.PathLength(taskAxis.Points);
            double error = Math.Abs(gestureLength - taskAxisLength);
            if (computeAsPercentage)
                error /= taskAxisLength;
            return error;
        }

        /// <summary>
        /// Size Error (SzE) is defined 
        /// as the absolute difference in area size from the task axis.
        /// </summary>
        public static double SizeError(Gesture gesture, CentroidGesture taskAxis, bool computeAsPercentage = false)
        {
            double gestureArea = PointSet.BoundingBox(gesture.Points).Area;
            double taskAxisArea = PointSet.BoundingBox(taskAxis.Points).Area;
            double error = Math.Abs(gestureArea - taskAxisArea);
            if (computeAsPercentage)
                error /= taskAxisArea;
            return error;
        }

        /// <summary>
        /// Bending Error (BE) is defined as the average of absolute differences in
        /// turning angle of each gesture point to the task axis.
        /// </summary>
        public static double BendingError(Gesture gesture, CentroidGesture taskAxis)
        {
            double[] errors = ComputeLocalBendingErrors(gesture, taskAxis);
            return ComputeAverage(errors);
        }

        /// <summary>
        /// Bending Variability (BV) is defined
        /// as the standard deviation of differences in turning angle of each point to the task axis.
        /// </summary>
        public static double BendingVariability(Gesture gesture, CentroidGesture taskAxis)
        {
            double[] errors = ComputeLocalBendingErrors(gesture, taskAxis);
            return ComputeStdev(errors);
        }

        private static double[] ComputeLocalBendingErrors(Gesture gesture, CentroidGesture taskAxis)
        {
            // align gesture points with the task axis
            Point[] taskAxisPts = taskAxis.GetPoints();
            Point[] gesturePts = taskAxis.AlignGesture(gesture);

            // compute errors
            double[] kTaskAxis = ComputeTurningAngleArray(taskAxisPts);
            double[] kGesture = ComputeTurningAngleArray(gesturePts);
            double[] errors = new double[gesturePts.Length];
            for (int i = 0; i < gesturePts.Length; i++)
                errors[i] = Math.Abs(kGesture[i] - kTaskAxis[i]);
            return errors;
        }

        /// <summary>
        /// Returns an array with turning angle values for each point.
        /// </summary>
        private static double[] ComputeTurningAngleArray(Point[] points)
        {
            double[] k = new double[points.Length];
            int n = k.Length;
            int r = 1;
            for (int i = 0; i < n; i++)
            {
                if (i - r >= 0 && i + r < n)
                {
                    double angle = Measurements.Angle(
                        new Vector(points[i], points[i + r]),
                        new Vector(points[i - r], points[i])
                    );
                    if (angle > Math.PI)
                        angle = angle - 2.0 * Math.PI;
                    k[i] = angle;
                }
                else k[i] = 0;
            }
            return k;
        }

        #endregion

        #region 4 kinematic relative accuracy measures: TE, TV, VE, VV

        /// <summary>
        /// Time Error (TE) is defined as the absolute difference in articulation time.
        /// </summary>
        /// <returns></returns>
        public static double TimeError(Gesture gesture, CentroidGesture taskAxis, bool computeAsPercentage = false)
        {
            double taskAxisTime = ProductionTime(taskAxis);
            double gestureTime = ProductionTime(gesture);
            double error = Math.Abs(gestureTime - taskAxisTime);
            if (computeAsPercentage)
                error /= taskAxisTime;
            return error;
        }

        /// <summary>
        /// Time Variability (TV) is defined as the standard deviation of local times.
        /// </summary>
        public static double TimeVariability(Gesture gesture, CentroidGesture taskAxis)
        {
            // align gesture points with the task axis
            Point[] taskAxisPts = taskAxis.GetPoints();
            Point[] gesturePts = taskAxis.AlignGesture(gesture);

            // compute errors
            double[] errors = new double[gesturePts.Length];
            for (int i = 0; i < gesturePts.Length; i++)
                errors[i] = Math.Abs(gesturePts[i].T - taskAxisPts[i].T);
            return ComputeStdev(errors);
        }

        /// <summary>
        /// Speed Error (VE) is defined as the absolute difference in speed.
        /// </summary>
        /// <returns></returns>
        public static double VelocityError(Gesture gesture, CentroidGesture taskAxis)
        {
            double[] errors = ComputeLocalSpeedErrors(gesture, taskAxis);
            return ComputeAverage(errors);
        }

        /// <summary>
        /// Velocity Variability is defined as the standard deviation of local speed values.
        /// </summary>
        public static double VelocityVariability(Gesture gesture, CentroidGesture taskAxis)
        {
            double[] errors = ComputeLocalSpeedErrors(gesture, taskAxis);
            return ComputeStdev(errors);
        }

        private static double[] ComputeLocalSpeedErrors(Gesture gesture, CentroidGesture taskAxis)
        {
            // align gesture points with the task axis
            Point[] taskAxisPts = taskAxis.GetPoints();
            Point[] gesturePts = taskAxis.AlignGesture(gesture);

            // compute measure
            double[] vTaskAxis = ComputeSpeedArray(taskAxisPts);
            double[] vGesture = ComputeSpeedArray(gesturePts);
            double[] errors = new double[gesturePts.Length];
            for (int i = 0; i < gesturePts.Length; i++)
                errors[i] = Math.Abs(vGesture[i] - vTaskAxis[i]);
            return errors;
        }

        /// <summary>
        /// Gets the production time of the gesture articulation.
        /// </summary>
        private static double ProductionTime(Gesture gesture)
        {
            Point[] points = gesture.Points;
            if (points.Length <= 1) return 0;
            return points[points.Length - 1].T - points[0].T;
        }

        /// <summary>
        /// Returns an array with speed values for each point.
        /// </summary>
        private static double[] ComputeSpeedArray(Point[] points)
        {
            int n = points.Length;
            double[] v = new double[n];
            int r = 1;
            for (int i = 0; i < n; i++)
            {
                int index1 = Math.Max(0, i - r);
                int index2 = Math.Min(i + r, n - 1);
                double distance = PointSet.PathLength(points, index1, index2);
                double time = points[index2].T - points[index1].T;
                if (Math.Abs(time) < 1e-5)
                    v[i] = 0;
                else v[i] = distance / time;
            }
            return v;
        }

        #endregion

        #region 2 articulation relative accuracy measures: SkE, SkOE

        /// <summary>
        /// Stroke Error (SkE) is defined as the absolute difference in stroke count.
        /// </summary>
        public static double StrokeError(Gesture gesture, CentroidGesture taskAxis)
        {
            return Math.Abs(NumberOfStrokes(gesture) - NumberOfStrokes(taskAxis));
        }

        /// <summary>
        /// Returns the number of strokes of a gesture.
        /// Complexity: O(n)
        /// </summary>
        /// <param name="gesture"></param>
        /// <param name="type"></param>
        /// <returns></returns>
        private static int NumberOfStrokes(Gesture gesture)
        {
            Point[] points = gesture.Points;
            int numStrokes = points.Length > 0 ? 1 : 0;
            for (int i = 1; i < points.Length; i++)
                if (points[i].StrokeID != points[i - 1].StrokeID)
                    numStrokes++;
            return numStrokes;
        }

        /// <summary>
        /// Stroke Order Error (SkOE) is defined as the absolute difference between the $1 and $P distance costs.
        /// </summary>
        public static double StrokeOrderError(Gesture gesture, CentroidGesture taskAxis)
        {
            // align gesture points with the task axis
            Point[] taskAxisPts = taskAxis.GetPoints();

            // compute $1 cost measure
            Point[] gesturePts = taskAxis.AlignGesture(gesture, PointAlignmentType.Chronological);
            double _1DollarCost = 0;
            for (int i = 0; i < gesturePts.Length; i++)
                _1DollarCost += Measurements.EuclideanDistance(gesturePts[i], taskAxisPts[i]);

            // compute $P cost measure
            gesturePts = taskAxis.AlignGesture(gesture, PointAlignmentType.PointCloudBestMatch);
            double _PDollarCost = 0;
            for (int i = 0; i < gesturePts.Length; i++)
                _PDollarCost += Measurements.EuclideanDistance(gesturePts[i], taskAxisPts[i]);

            return Math.Abs(_1DollarCost - _PDollarCost);
        }

        #endregion

        #region utility functions: average, stdev

        /// <summary>
        /// Computes the average of a set of real values.
        /// </summary>
        private static double ComputeAverage(double[] values)
        {
            double avg = 0;
            foreach (double v in values)
                avg += v;
            avg /= values.Length;
            return avg;
        }

        /// <summary>
        /// Computes the standard deviation for a set of real values.
        /// </summary>
        private static double ComputeStdev(double[] values)
        {
            double avg = ComputeAverage(values);
            double stdev = 0;
            foreach (double v in values)
                stdev += (v - avg) * (v - avg);
            stdev = Math.Sqrt(stdev / (values.Length - 1));
            return stdev;
        }

        #endregion
    }
}