﻿/**
 * Relative Accuracy Measures for Stroke Gestures (.NET Framework 4.5 C# version)
 *
 * 	    Radu-Daniel Vatavu, Ph.D.
 *	    University Stefan cel Mare of Suceava
 *	    Suceava 720229, Romania
 *	    vatavu@eed.usv.ro
 *
 * 	    Lisa Anthony, Ph.D.
 *      UMBC
 *      Information Systems Department
 *      1000 Hilltop Circle
 *      Baltimore, MD 21250
 *      lanthony@umbc.edu
 *	
 *	    Jacob O. Wobbrock, Ph.D.
 * 	    The Information School
 *	    University of Washington
 *	    Seattle, WA 98195-2840
 *	    wobbrock@uw.edu
 *
 * The academic publication for the relative accuracy measures, and what should be 
 * used to cite it, is:
 *
 *	Vatavu, R.-D., Anthony, L. and Wobbrock, J.O. (2013).  
 *	  Relative accuracy measures for stroke gestures. 
 *	  In Proceedings of the 15th ACM on International Conference 
 *	  on Multimodal Interaction (ICMI '13). ACM, New York, NY, USA, 279-286
 *	  DOI=10.1145/2522848.2522875 http://doi.acm.org/10.1145/2522848.2522875
 *
 * This software is distributed under the "New BSD License" agreement:
 *
 * Copyright (c) 2013, Radu-Daniel Vatavu, Lisa Anthony, and 
 * Jacob O. Wobbrock. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the names of the University Stefan cel Mare of Suceava, 
 *	    University of Washington, nor UMBC, nor the names of its contributors 
 *	    may be used to endorse or promote products derived from this software 
 *	    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Radu-Daniel Vatavu OR Lisa Anthony
 * OR Jacob O. Wobbrock BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
**/
using System;
using System.Collections.Generic;
using System.IO;
using System.Windows.Forms;
using RelativeAccuracyMeasures;

namespace GREAT
{
    public partial class FormMain : Form
    {
        private Dictionary<string, List<Gesture>> gestureSet = null;

        public FormMain()
        {
            InitializeComponent();

            // adjust the sizes of the columns in the data grid view
            resultsGrid.Columns[0].Width = 110;
            resultsGrid.Columns[1].Width = 40;
            resultsGrid.Columns[1].DefaultCellStyle.Alignment = DataGridViewContentAlignment.MiddleRight;
            for (int i = 2; i < resultsGrid.Columns.Count; i++)
            {
                resultsGrid.Columns[i].Width = 70;
                resultsGrid.Columns[i].DefaultCellStyle.Alignment = DataGridViewContentAlignment.MiddleRight;
            }

            // disable some menu choices
            computeRelativeAccuracyMeasuresToolStripMenuItem.Enabled = false;
            btnExport.Enabled = false;

            rbChronologicalOrder.Focus();
        }

        private void loadGesturesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            folderBrowserDialog.SelectedPath = Application.StartupPath;
            if (folderBrowserDialog.ShowDialog() == System.Windows.Forms.DialogResult.OK)
            {
                // load all gestures located under the selected folder
                gestureSet = new Dictionary<string, List<Gesture>>();
                string[] xmlFiles = System.IO.Directory.GetFiles(folderBrowserDialog.SelectedPath, "*.xml");
                int numGestures = 0;
                foreach (string xmlFile in xmlFiles)
                {
                    Gesture gesture = GestureIO.ReadGesture(xmlFile);
                    if (!gestureSet.ContainsKey(gesture.Name))
                        gestureSet.Add(gesture.Name, new List<Gesture>());
                    gestureSet[gesture.Name].Add(gesture);
                    numGestures++;
                }
                
                // inform the user
                string text = String.Format(
                    "Dataset: {0}\r\nA number of {1} gesture samples were successfully loaded for {2} distinct gesture types:\r\n", 
                    folderBrowserDialog.SelectedPath,
                    numGestures,
                    gestureSet.Keys.Count
                );
                foreach (string type in gestureSet.Keys)
                    text += type + "; ";
                tbGestureSetInfo.Text = text;

                // enable menu option to compute measures
                computeRelativeAccuracyMeasuresToolStripMenuItem.Enabled = true;
            }
        }

        private void computeRelativeAccuracyMeasuresToolStripMenuItem_Click(object sender, EventArgs e)
        {
            // parameters for computing the task axis
            PointAlignmentType alignmentType = rbChronologicalOrder.Checked ? PointAlignmentType.Chronological : PointAlignmentType.PointCloudBestMatch;
            bool useNearestNeighbor = rbNearestNeighbor.Checked;

            resultsGrid.Rows.Clear();
            foreach (KeyValuePair<string, List<Gesture>> gesturesRecord in gestureSet)
            {
                CentroidGesture taskAxis = new CentroidGesture(gesturesRecord.Value.ToArray(), alignmentType, useNearestNeighbor);
                for (int i = 0; i < gesturesRecord.Value.Count; i++)
                {
                    int row = resultsGrid.Rows.Add();
                    resultsGrid.Rows[row].Cells[0].Value = gesturesRecord.Key;
                    resultsGrid.Rows[row].Cells[1].Value = i + 1;
                    resultsGrid.Rows[row].Cells[2].Value = String.Format("{0:.000}", RelativeAccuracy.ShapeError(gesturesRecord.Value[i], taskAxis));
                    resultsGrid.Rows[row].Cells[3].Value = String.Format("{0:.000}", RelativeAccuracy.ShapeVariability(gesturesRecord.Value[i], taskAxis));
                    resultsGrid.Rows[row].Cells[4].Value = String.Format("{0:.000}", RelativeAccuracy.LengthError(gesturesRecord.Value[i], taskAxis));
                    resultsGrid.Rows[row].Cells[5].Value = String.Format("{0:.000}", RelativeAccuracy.SizeError(gesturesRecord.Value[i], taskAxis));
                    resultsGrid.Rows[row].Cells[6].Value = String.Format("{0:.000}", RelativeAccuracy.BendingError(gesturesRecord.Value[i], taskAxis));
                    resultsGrid.Rows[row].Cells[7].Value = String.Format("{0:.000}", RelativeAccuracy.BendingVariability(gesturesRecord.Value[i], taskAxis));
                    resultsGrid.Rows[row].Cells[8].Value = String.Format("{0:.000}", RelativeAccuracy.TimeError(gesturesRecord.Value[i], taskAxis));
                    resultsGrid.Rows[row].Cells[9].Value = String.Format("{0:.000}", RelativeAccuracy.TimeVariability(gesturesRecord.Value[i], taskAxis));
                    resultsGrid.Rows[row].Cells[10].Value = String.Format("{0:.000}", RelativeAccuracy.VelocityError(gesturesRecord.Value[i], taskAxis));
                    resultsGrid.Rows[row].Cells[11].Value = String.Format("{0:.000}", RelativeAccuracy.VelocityVariability(gesturesRecord.Value[i], taskAxis));
                    resultsGrid.Rows[row].Cells[12].Value = String.Format("{0}", (int)RelativeAccuracy.StrokeError(gesturesRecord.Value[i], taskAxis));
                    resultsGrid.Rows[row].Cells[13].Value = String.Format("{0:.000}", RelativeAccuracy.StrokeOrderError(gesturesRecord.Value[i], taskAxis));
                }
            }

            // enable exporting of results
            btnExport.Enabled = true;
        }

        private void btnExport_Click(object sender, EventArgs e)
        {
            if (saveFileDialog.ShowDialog() == System.Windows.Forms.DialogResult.OK)
            {
                using (StreamWriter writer = new StreamWriter(saveFileDialog.FileName))
                {
                    writer.WriteLine("GestureType,No,ShE,ShV,LE,SzE,BE,BV,TE,TV,VE,VV,SkE,SkOE");
                    foreach (DataGridViewRow row in resultsGrid.Rows)
                    {
                        for (int i = 0; i < row.Cells.Count; i++)
                            writer.Write("{0}{1}", row.Cells[i].Value, i == row.Cells.Count - 1 ? "" : ",");
                        writer.WriteLine();
                    }
                }
            }
        }
    }
}
